import mythic_container
from mythic_container.logging import logger
from .send_mythic_rpc_callbacktoken_create import Token

MYTHIC_RPC_TOKEN_REMOVE = "mythic_rpc_token_remove"



class MythicRPCTokenRemoveMessage:
    def __init__(self,
                 TaskID: int,
                 Tokens: list[Token],
                 **kwargs):
        self.TaskID = TaskID
        self.Tokens = Tokens
        for k, v in kwargs.items():
            logger.info(f"Unknown kwarg {k} - {v}")

    def to_json(self):
        return {
            "task_id": self.TaskID,
            "tokens": [x.to_json() for x in self.Tokens]
        }


class MythicRPCTokenRemoveMessageResponse:
    def __init__(self,
                 success: bool = False,
                 error: str = "",
                 **kwargs):
        self.Success = success
        self.Error = error
        for k, v in kwargs.items():
            logger.info(f"Unknown kwarg {k} - {v}")


async def SendMythicRPCTokenRemove(
        msg: MythicRPCTokenRemoveMessage) -> MythicRPCTokenRemoveMessageResponse:
    for t in msg.Tokens:
        t.Action = "remove"
    response = await mythic_container.RabbitmqConnection.SendRPCDictMessage(queue=MYTHIC_RPC_TOKEN_REMOVE,
                                                                            body=msg.to_json())
    return MythicRPCTokenRemoveMessageResponse(**response)



import mythic_container
from mythic_container.logging import logger

MYTHIC_RPC_PAYLOAD_REMOVE_COMMAND = "mythic_rpc_payload_remove_command"


class MythicRPCPayloadRemoveCommandMessage:
    def __init__(self,
                 PayloadUUID: str,
                 Commands: list[str],
                 **kwargs):
        self.PayloadUUID = PayloadUUID
        self.Commands = Commands
        for k, v in kwargs.items():
            logger.info(f"Unknown kwarg {k} - {v}")

    def to_json(self):
        return {
            "payload_uuid": self.PayloadUUID,
            "commands": self.Commands
        }


class MythicRPCPayloadRemoveCommandMessageResponse:
    def __init__(self,
                 success: bool = False,
                 error: str = "",
                 **kwargs):
        self.Success = success
        self.Error = error
        for k, v in kwargs.items():
            logger.info(f"Unknown kwarg {k} - {v}")


async def SendMythicRPCPayloadRemoveCommand(
        msg: MythicRPCPayloadRemoveCommandMessage) -> MythicRPCPayloadRemoveCommandMessageResponse:
    response = await mythic_container.RabbitmqConnection.SendRPCDictMessage(queue=MYTHIC_RPC_PAYLOAD_REMOVE_COMMAND,
                                                                            body=msg.to_json())
    return MythicRPCPayloadRemoveCommandMessageResponse(**response)



from .logging import logger
import ujson
import mythic_container.C2ProfileBase
import psutil
import os
import subprocess
import asyncio
import traceback


def kill(proc_pid):
    try:
        target_processes = psutil.Process(proc_pid)
        for proc in target_processes.children(recursive=True):
            proc.kill()
        target_processes.kill()
    except psutil.NoSuchProcess:
        pass
    except Exception as e:
        logger.exception(f"[-] Failed to kill process: {e}")


async def deal_with_stdout(c2_profile: str) -> str:
    output = ""
    lines = 0
    while True:
        try:
            line = await asyncio.wait_for(
                mythic_container.C2ProfileBase.runningServers[c2_profile]["process"].stdout.readline(), timeout=3.0)
            output += line.decode()
            lines += 1
            if lines > 100:
                break
        except asyncio.TimeoutError:
            break
        except Exception as e:
            logger.exception(f"hit exception trying to get server output: {traceback.format_exc()}")
            return output + traceback.format_exc()
    return output


async def opsecChecks(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if callable(c2.opsec):
                    try:
                        result = await c2.opsec(mythic_container.C2ProfileBase.C2OPSECMessage(**msgDict))
                    except Exception as callEx:
                        logger.exception(
                            f"Failed to call opsec for {c2.name}")
                        response = mythic_container.C2ProfileBase.C2OPSECMessageResponse(
                            Success=False,
                            Error=f"Failed to call opsec function: {traceback.format_exc()}\n{callEx}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    if result is None:
                        response = mythic_container.C2ProfileBase.C2OPSECMessageResponse(
                            Success=False,
                            Error=f"Failed to call opsec function: No result returned"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    elif isinstance(result, dict):
                        response = mythic_container.C2ProfileBase.C2OPSECMessageResponse(**result)
                        return ujson.dumps(response).encode()
                    elif isinstance(result, mythic_container.C2ProfileBase.C2OPSECMessageResponse):
                        return ujson.dumps(result.to_json()).encode()
                    else:
                        response = mythic_container.C2ProfileBase.C2OPSECMessageResponse(
                            Success=False,
                            Error=f"unknown result type from function: {result}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                else:
                    logger.error(f"opsec function for {c2.name} isn't callable")
                    response = mythic_container.C2ProfileBase.C2OPSECMessageResponse(
                        Success=False,
                        Error=f"opsec function for {c2.name} isn't callable"
                    )
                    return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2OPSECMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call opsec function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def configChecks(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if callable(c2.opsec):
                    try:
                        result = await c2.config_check(mythic_container.C2ProfileBase.C2ConfigCheckMessage(**msgDict))
                    except Exception as callEx:
                        logger.exception(
                            f"Failed to call config check for {c2.name}")
                        response = mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse(
                            Success=False,
                            Error=f"Failed to call config check function: {traceback.format_exc()}\n{callEx}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    if result is None:
                        response = mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse(
                            Success=False,
                            Error=f"Failed to call config check function: No result returned"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    elif isinstance(result, dict):
                        response = mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse(**result)
                        return ujson.dumps(response).encode()
                    elif isinstance(result, mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse):
                        return ujson.dumps(result.to_json()).encode()
                    else:
                        response = mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse(
                            Success=False,
                            Error=f"unknown result type from function: {result}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                else:
                    logger.error(f"config check function for {c2.name} isn't callable")
                    response = mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse(
                        Success=False,
                        Error=f"config check function for {c2.name} isn't callable"
                    )
                    return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2ConfigCheckMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call config check function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def getIOC(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if callable(c2.get_ioc):
                    try:
                        result = await c2.config_check(mythic_container.C2ProfileBase.C2GetIOCMessage(**msgDict))
                    except Exception as callEx:
                        logger.exception(
                            f"Failed to call get_ioc for {c2.name}")
                        response = mythic_container.C2ProfileBase.C2GetIOCMessageResponse(
                            Success=False,
                            Error=f"Failed to call config check function: {traceback.format_exc()}\n{callEx}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    if result is None:
                        response = mythic_container.C2ProfileBase.C2GetIOCMessageResponse(
                            Success=False,
                            Error=f"Failed to call config check function: No result returned"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    elif isinstance(result, dict):
                        response = mythic_container.C2ProfileBase.C2GetIOCMessageResponse(**result)
                        return ujson.dumps(response).encode()
                    elif isinstance(result, mythic_container.C2ProfileBase.C2GetIOCMessageResponse):
                        return ujson.dumps(result.to_json()).encode()
                    else:
                        response = mythic_container.C2ProfileBase.C2GetIOCMessageResponse(
                            Success=False,
                            Error=f"unknown result type from function: {result}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                else:
                    logger.error(f"get_ioc function for {c2.name} isn't callable")
                    response = mythic_container.C2ProfileBase.C2GetIOCMessageResponse(
                        Success=False,
                        Error=f"get ioc function for {c2.name} isn't callable"
                    )
                    return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2GetIOCMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call get_ioc function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def sampleMessage(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if callable(c2.sample_message):
                    try:
                        result = await c2.config_check(mythic_container.C2ProfileBase.C2SampleMessageMessage(**msgDict))
                    except Exception as callEx:
                        logger.exception(
                            f"Failed to call get_ioc for {c2.name}")
                        response = mythic_container.C2ProfileBase.C2SampleMessageMessageResponse(
                            Success=False,
                            Error=f"Failed to call sample_message function: {traceback.format_exc()}\n{callEx}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    if result is None:
                        response = mythic_container.C2ProfileBase.C2SampleMessageMessageResponse(
                            Success=False,
                            Error=f"Failed to call sample message function: No result returned"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    elif isinstance(result, dict):
                        response = mythic_container.C2ProfileBase.C2SampleMessageMessageResponse(**result)
                        return ujson.dumps(response).encode()
                    elif isinstance(result, mythic_container.C2ProfileBase.C2SampleMessageMessageResponse):
                        return ujson.dumps(result.to_json()).encode()
                    else:
                        response = mythic_container.C2ProfileBase.C2SampleMessageMessageResponse(
                            Success=False,
                            Error=f"unknown result type from function: {result}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                else:
                    logger.error(f"sample_message function for {c2.name} isn't callable")
                    response = mythic_container.C2ProfileBase.C2SampleMessageMessageResponse(
                        Success=False,
                        Error=f"sample message function for {c2.name} isn't callable"
                    )
                    return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2SampleMessageMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call sample_message function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def getDebugOutput(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if mythic_container.C2ProfileBase.runningServers[c2.name]["process"] is not None:
                    if mythic_container.C2ProfileBase.runningServers[c2.name]["process"].returncode is None:
                        running = True
                    else:
                        running = False
                        try:
                            kill(mythic_container.C2ProfileBase.runningServers[c2.name]["process"].pid)
                        except Exception as e:
                            pass
                    response = mythic_container.C2ProfileBase.C2GetDebugOutputMessageResponse(
                        Success=True, Error="", Message=f"{await deal_with_stdout(c2.name)}",
                        InternalServerRunning=running
                    )
                    return ujson.dumps(response.to_json()).encode()
                else:
                    response = mythic_container.C2ProfileBase.C2GetDebugOutputMessageResponse(
                        Success=True, Error="", Message="Server not running",
                        InternalServerRunning=False
                    )
                    return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2GetDebugOutputMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
            InternalServerRunning=False
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2GetDebugOutputMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call get debug output function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def getFile(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                response = mythic_container.C2ProfileBase.C2GetFileMessageResponse(Success=False)
                try:
                    inputMsg = mythic_container.C2ProfileBase.C2GetFileMessage(**msgDict)
                    path = c2.server_folder_path / inputMsg.Filename
                    path = path.resolve()
                    if c2.server_folder_path.resolve() not in path.parents:
                        response.Success = False
                        response.Error = "Attempt to access outside of the c2 server folder"
                    else:
                        file_data = open(path, "rb").read()
                        response.Success = True
                        response.Message = file_data
                except Exception as e:
                    response.Error = f"{traceback.format_exc()}\n{e}"
                return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2GetFileMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2GetFileMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call get debug output function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def getRedirectorRules(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if callable(c2.redirect_rules):
                    try:
                        result = await c2.redirect_rules(
                            mythic_container.C2ProfileBase.C2GetRedirectorRulesMessage(**msgDict))
                    except Exception as callEx:
                        logger.exception(
                            f"Failed to call get redirector rules for {c2.name}")
                        response = mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse(
                            Success=False,
                            Error=f"Failed to call get redirector rules function: {traceback.format_exc()}\n{callEx}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    if result is None:
                        response = mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse(
                            Success=False,
                            Error=f"Failed to call get redirector rules function: No result returned"
                        )
                        return ujson.dumps(response.to_json()).encode()
                    elif isinstance(result, dict):
                        response = mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse(**result)
                        return ujson.dumps(response).encode()
                    elif isinstance(result, mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse):
                        return ujson.dumps(result.to_json()).encode()
                    else:
                        response = mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse(
                            Success=False,
                            Error=f"unknown result type from function: {result}"
                        )
                        return ujson.dumps(response.to_json()).encode()
                else:
                    logger.error(f"get redirector rules function for {c2.name} isn't callable")
                    response = mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse(
                        Success=False,
                        Error=f"get redirector rules function for {c2.name} isn't callable"
                    )
                    return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2GetRedirectorRulesMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call get redirector rules function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def listFile(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                response = mythic_container.C2ProfileBase.C2ListFileMessageResponse(Success=False)
                try:
                    path = c2.server_folder_path
                    files = os.listdir(path)
                    files = [f for f in files if os.path.isfile(path / f)]
                    response.Files = files
                    response.Success = True
                except Exception as e:
                    response.Error = f"{traceback.format_exc()}\n{e}"
                return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2ListFileMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2ListFileMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call list file function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def removeFile(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                response = mythic_container.C2ProfileBase.C2RemoveFileMessageResponse(Success=False)
                try:
                    inputMsg = mythic_container.C2ProfileBase.C2RemoveFileMessage(**msgDict)
                    path = c2.server_folder_path / inputMsg.Filename
                    path = path.resolve()
                    if c2.server_folder_path.resolve() not in path.parents:
                        response.Success = False
                        response.Error = "Attempt to access outside of the c2 server folder"
                    else:
                        os.remove(path)
                        response.Success = True
                except Exception as e:
                    response.Error = f"{traceback.format_exc()}\n{e}"
                return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2GetFileMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2GetFileMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call get debug output function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def startServerBinary(
        c2: mythic_container.C2ProfileBase.C2Profile) -> mythic_container.C2ProfileBase.C2StartServerMessageResponse:
    try:
        if c2.name not in mythic_container.C2ProfileBase.runningServers:
            mythic_container.C2ProfileBase.runningServers[c2.name] = {
                "process": None
            }
        path = c2.server_binary_path.resolve()
        cwd = c2.server_folder_path.resolve()
        os.chmod(path, mode=0o777)
        process = await asyncio.create_subprocess_shell(cmd=str(path), stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                                        shell=True, cwd=str(cwd), env=os.environ.copy())
        mythic_container.C2ProfileBase.runningServers[c2.name]["process"] = process
        await asyncio.sleep(3)
        if process.returncode is not None:
            # this means something went wrong and the process is dead
            return mythic_container.C2ProfileBase.C2StartServerMessageResponse(
                Success=False,
                Error=f"Failed to start\nOutput:\n{await deal_with_stdout(c2.name)}",
                InternalServerRunning=False
            )
        else:
            return mythic_container.C2ProfileBase.C2StartServerMessageResponse(
                Success=True,
                Message=f"Started with pid: {process.pid}...\nOutput:{await deal_with_stdout(c2.name)}",
                InternalServerRunning=True
            )
    except Exception as e:
        return mythic_container.C2ProfileBase.C2StartServerMessageResponse(
            Success=False,
            Error=f"Failed to start server: {traceback.format_exc()}\n{e}",
            Message=f"Failed to start server: {traceback.format_exc()}\n{e}",
            InternalServerRunning=False
        )


async def startServer(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if c2.name in mythic_container.C2ProfileBase.runningServers:
                    # we've at least started this before, so we can check some keys
                    if mythic_container.C2ProfileBase.runningServers[c2.name]["process"] is None:
                        # process isn't running, so we need to start it
                        response = await startServerBinary(c2)
                        return ujson.dumps(response.to_json()).encode()

                    elif mythic_container.C2ProfileBase.runningServers[c2.name]["process"].returncode is None:
                        response = mythic_container.C2ProfileBase.C2StartServerMessageResponse(
                            Success=True, Error="", Message=await deal_with_stdout(c2.name),
                            InternalServerRunning=True
                        )
                        return ujson.dumps(response.to_json()).encode()
                    else:
                        try:
                            kill(mythic_container.C2ProfileBase.runningServers[c2.name]["process"].pid)
                        except Exception as e:
                            pass
                        # now try to start it again
                        response = await startServerBinary(c2)
                        return ujson.dumps(response.to_json()).encode()
                else:
                    # just means we've never started it before, so start it now
                    response = await startServerBinary(c2)
                    return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2StartServerMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
            InternalServerRunning=False
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2StartServerMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call server start function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def stopServer(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                if c2.name in mythic_container.C2ProfileBase.runningServers:
                    # we've at least started this before, so we can check some keys
                    if mythic_container.C2ProfileBase.runningServers[c2.name]["process"] is None:
                        # process isn't running, so we need to start it
                        response = mythic_container.C2ProfileBase.C2StopServerMessageResponse(
                            Success=True, Error="", Message=f"The server is not running",
                            InternalServerRunning=False
                        )
                        return ujson.dumps(response.to_json()).encode()

                    elif mythic_container.C2ProfileBase.runningServers[c2.name]["process"].returncode is not None:
                        # the process was running, but died at some point before we tried to stop it
                        response = mythic_container.C2ProfileBase.C2StartServerMessageResponse(
                            Success=True, Error="", Message=f"{await deal_with_stdout(c2.name)}",
                            InternalServerRunning=False
                        )
                        return ujson.dumps(response.to_json()).encode()
                    else:
                        try:
                            kill(mythic_container.C2ProfileBase.runningServers[c2.name]["process"].pid)
                        except Exception as e:
                            pass
                        # now try to start it again
                        await asyncio.sleep(3)
                        response = mythic_container.C2ProfileBase.C2StopServerMessageResponse(
                            Success=True, Error="",
                            Message=f"Stopped server:\nOutput:{await deal_with_stdout(c2.name)}",
                            InternalServerRunning=False
                        )
                        mythic_container.C2ProfileBase.runningServers[c2.name]["output"] = ""
                        return ujson.dumps(response.to_json()).encode()
                else:
                    # just means we've never started it before, so start it now
                    response = mythic_container.C2ProfileBase.C2StopServerMessageResponse(
                        Success=True, Error="", Message=f"The server was never started",
                        InternalServerRunning=False
                    )
                    return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2StartServerMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
            InternalServerRunning=False
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2StartServerMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call server start function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def writeFile(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                response = mythic_container.C2ProfileBase.C2WriteFileMessageResponse(Success=False)
                try:
                    inputMsg = mythic_container.C2ProfileBase.C2WriteFileMessage(**msgDict)
                    path = c2.server_folder_path / inputMsg.Filename
                    path = path.resolve()
                    if c2.server_folder_path.resolve() not in path.parents:
                        response.Success = False
                        response.Error = "Attempt to write outside of the c2 server folder"
                    else:
                        with open(path, "wb") as f:
                            f.write(inputMsg.Contents)
                        response.Success = True
                        response.Message = "Successfully wrote file"
                except Exception as e:
                    logger.exception(f"[-] Failed to write to file: {e}")
                    response.Error = f"{traceback.format_exc()}\n{e}"
                return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2WriteFileMessageResponse(
            Success=False, Error="Failed to find that c2 profile",
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        logger.exception(f"[-] Failed to write to file with exception: {e}")
        response = mythic_container.C2ProfileBase.C2WriteFileMessageResponse(
            Success=False,
            Error=f"Hit exception trying to call write file function function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def customRPCFunction(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        incomingMessage = mythic_container.C2ProfileBase.C2OtherServiceRPCMessage(**msgDict)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == incomingMessage.ServiceName:
                if c2.custom_rpc_functions is not None:
                    for key, f in c2.custom_rpc_functions.items():
                        if key == incomingMessage.ServiceRPCFunction:
                            response = await f(incomingMessage)
                            return ujson.dumps(response.to_json()).encode()
                    response = mythic_container.C2ProfileBase.C2OtherServiceRPCMessageResponse(
                        Success=False,
                        Error=f"Failed to find function {incomingMessage.ServiceRPCFunction} in custom rpc functions for {c2.name}"
                    )
                    return ujson.dumps(response.to_json()).encode()
                else:
                    response = mythic_container.C2ProfileBase.C2OtherServiceRPCMessageResponse(
                        Success=False,
                        Error=f"No custom rpc functions defined for payload {c2.name}"
                    )
                    return ujson.dumps(response.to_json()).encode()
        response = mythic_container.C2ProfileBase.C2OtherServiceRPCMessageResponse(
            Success=False,
            Error=f"Failed to find service: {incomingMessage.ServiceName}"
        )
        return ujson.dumps(response.to_json()).encode()
    except Exception as e:
        response = mythic_container.C2ProfileBase.C2OtherServiceRPCMessageResponse(
            Success=False,
            Error=f"Failed to call custom RPC function: {traceback.format_exc()}\n{e}"
        )
        return ujson.dumps(response.to_json()).encode()


async def reSyncC2Profile(msg: bytes) -> bytes:
    try:
        msgDict = ujson.loads(msg)
        for name, c2 in mythic_container.C2ProfileBase.c2Profiles.items():
            if c2.name == msgDict["c2_profile_name"]:
                await mythic_container.mythic_service.syncC2ProfileData(c2)
                return ujson.dumps({"success": True}).encode()
        return ujson.dumps({"success": False, "error": "Failed to find c2 profile"}).encode()
    except Exception as e:
        logger.exception(f"Failed to re-sync c2 profile: {e}")
        return ujson.dumps({"success": False, "error": f"Failed to sync: {traceback.format_exc()}\n{e}"}).encode()



import ujson
from .logging import logger
from .WebhookBase import *


async def new_startup(msg: bytes) -> None:
    try:
        msgDict = ujson.loads(msg)
        webhookServices = Webhook.__subclasses__()
        for cls in webhookServices:
            webhook = cls()
            if webhook.new_startup is not None and callable(webhook.new_startup):
                await webhook.new_startup(WebhookMessage(**msgDict))
            else:
                logger.error(f"No valid function for new_startup from webhook")
    except Exception as e:
        logger.exception(f"Failed to execute new_startup webhook: {e}")


async def new_callback(msg: bytes) -> None:
    try:
        msgDict = ujson.loads(msg)
        webhookServices = Webhook.__subclasses__()
        for cls in webhookServices:
            webhook = cls()
            if webhook.new_callback is not None and callable(webhook.new_callback):
                await webhook.new_callback(WebhookMessage(**msgDict))
            else:
                logger.error(f"No valid function for new_callback from webhook")
    except Exception as e:
        logger.exception(f"Failed to execute new_callback webhook: {e}")


async def new_feedback(msg: bytes) -> None:
    try:
        msgDict = ujson.loads(msg)
        webhookServices = Webhook.__subclasses__()
        for cls in webhookServices:
            webhook = cls()
            if webhook.new_feedback is not None and callable(webhook.new_feedback):
                await webhook.new_feedback(WebhookMessage(**msgDict))
            else:
                logger.error(f"No valid function for new_feedback from webhook")
    except Exception as e:
        logger.exception(f"Failed to execute new_feedback webhook: {e}")


async def new_alert(msg: bytes) -> None:
    try:
        msgDict = ujson.loads(msg)
        webhookServices = Webhook.__subclasses__()
        for cls in webhookServices:
            webhook = cls()
            if webhook.new_alert is not None and callable(webhook.new_alert):
                await webhook.new_alert(WebhookMessage(**msgDict))
            else:
                logger.error(f"No valid function for new_alert from webhook")
    except Exception as e:
        logger.exception(f"Failed to execute new_alert webhook: {e}")


async def new_custom(msg: bytes) -> None:
    try:
        msgDict = ujson.loads(msg)
        webhookServices = Webhook.__subclasses__()
        for cls in webhookServices:
            webhook = cls()
            if webhook.new_custom is not None and callable(webhook.new_custom):
                await webhook.new_custom(WebhookMessage(**msgDict))
            else:
                logger.error(f"No valid function for new_custom from webhook")
    except Exception as e:
        logger.exception(f"Failed to execute new_custom webhook: {e}")


import concurrent.futures

import aiormq.exceptions

from .logging import logger
import aio_pika
import asyncio
import mythic_container
from .config import settings
from collections.abc import Callable, Awaitable
from typing import Coroutine
import ujson
import asyncio
from functools import partial
import uuid
from typing import MutableMapping


mutex = asyncio.Lock()

failedConnectRetryDelay = 5
failedConnectTimeout = 1


async def messageProcessThread(message: aio_pika.abc.AbstractIncomingMessage,
                               trueFunction: Callable[[bytes], Awaitable[None]]) -> None:
    try:
        await trueFunction(message.body)
    except Exception as d:
        logger.exception(f"inner error: {d}")


async def directExchangeCallback(message: aio_pika.abc.AbstractIncomingMessage,
                                 trueFunction: Callable[[bytes], Awaitable[None]]) -> None:
    # run async supplied function as background thread
    logger.debug(f"Got direct call to {message.routing_key}")
    async with message.process() as messageContext:
        # _thread = Thread(target=asyncio.run,
        #                 args=(messageProcessThread(message=messageContext, trueFunction=trueFunction),))
        # _thread.start()  # start thread
        await messageProcessThread(message=messageContext, trueFunction=trueFunction)


async def messageProcessRPCThread(message: aio_pika.abc.AbstractIncomingMessage,
                                  trueFunction: Callable[[bytes], Awaitable[bytes]]) -> bytes:
    try:
        response = await trueFunction(message.body)
        #logger.info(f"rpc response: {response}\nrequest: {message.body}")
        return response
    except Exception as d:
        logger.exception(f"rpc inner error: {d}")
        return f"rpc error: {d}".encode()


async def rpcExchangeCallback(message: aio_pika.abc.AbstractIncomingMessage,
                              trueFunction: Callable[[bytes], Awaitable[any]]) -> None:
    # run async supplied function as background thread
    async with message.process() as messageContext:
        # _thread = Thread(target=asyncio.run,
        #                 args=(messageProcessThread(message=messageContext, trueFunction=trueFunction),))
        # _thread.start()  # start thread
        logger.debug(f"Got RPC call to {message.routing_key}")
        response = await messageProcessRPCThread(message=messageContext, trueFunction=trueFunction)
        await mythic_container.RabbitmqConnection.ReplyMessage(response=response, message=messageContext)


class rabbitmqConnectionClass:
    conn: aio_pika.robust_connection.AbstractRobustConnection = None
    futures: MutableMapping[str, asyncio.Future] = {}

    def __init__(self):
        pass

    async def async_init(self):
        self.conn = await self.GetConnection()
        return self

    def __await__(self):
        return self.async_init().__await__()

    async def GetConnection(self) -> aio_pika.robust_connection.AbstractRobustConnection:
        async with mutex:
            while True:
                try:
                    if self.conn is not None and not self.conn.is_closed:
                        return self.conn
                    logger.info("[*] Trying to connect to rabbitmq at: "
                                 + settings.get("rabbitmq_host", "127.0.0.1")
                                 + ":"
                                 + str(settings.get("rabbitmq_port", 5672))
                                 )
                    logger.debug(f"connecting with password: {settings.get('rabbitmq_password', 'rabbitmq_password')}...")
                    self.conn = await aio_pika.connect_robust(
                        host=settings.get("rabbitmq_host", "127.0.0.1"),
                        port=settings.get("rabbitmq_port", 5672),
                        login="mythic_user",
                        password=settings.get("rabbitmq_password", "rabbitmq_password"),
                        virtualhost="mythic_vhost",
                        timeout=failedConnectTimeout,
                    )
                    logger.info("[+] Successfully connected to rabbitmq")
                    return self.conn
                except Exception as e:
                    logger.error(f"[-] Failed to connect to rabbitmq: {e}")
                    await asyncio.sleep(failedConnectRetryDelay)

    async def SendMessage(self, queue: str, body: bytes):
        while True:
            try:
                connection = await self.GetConnection()
                async with connection.channel(publisher_confirms=True,
                                              on_return_raises=True) as chan:
                    exchange = await chan.declare_exchange("mythic_exchange",
                                                           durable=True,
                                                           auto_delete=True)
                    message = aio_pika.Message(body=body,
                                               content_type="application/json")
                    await exchange.publish(
                        message=message,
                        routing_key=queue,
                        timeout=failedConnectRetryDelay,
                        mandatory=True,
                        immediate=False)
                    return
            except Exception as e:
                logger.exception(f"[-] failed to send message: {e}")
                return

    async def SendDictDirectMessage(self, queue: str, body: dict) -> None:
        #logger.debug(f"Sending Direct msg to {queue}: {body}")
        return await self.SendMessage(queue=queue, body=ujson.dumps(body).encode())

    async def SendRPCMessage(self, queue: str, body: bytes) -> dict:
        future = asyncio.get_event_loop().create_future()
        logger.debug(f"Sending RPC message to {queue}")
        try:
            correlation_id = str(uuid.uuid4())
            self.futures[correlation_id] = future
            while True:
                connection = await self.GetConnection()
                async with connection.channel(on_return_raises=True) as chan:
                    exchange = await chan.declare_exchange("mythic_exchange",
                                                           durable=True,
                                                           auto_delete=True)
                    callback_queue = await chan.declare_queue(name="amq.rabbitmq.reply-to",)
                    await callback_queue.consume(self.on_response,
                                                 no_ack=True)
                    message = aio_pika.Message(body=body,
                                               content_type="application/json",
                                               reply_to=callback_queue.name,
                                               correlation_id=correlation_id)
                    # make sure the queue exists first before we try to send to it
                    await exchange.publish(
                        message=message,
                        routing_key=queue,
                        mandatory=True,
                        immediate=False,
                        timeout=20
                    )
                    result = await future
                    return result
        except Exception as e:
            logger.error(f"[-] failed to send rpc message to {queue}: {e}")
            future.set_result({})
            return {}

    def on_response(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        try:
            #logger.debug(f"got on_response for correlation_id: {message.correlation_id}")
            if message.correlation_id is None:
                logger.error(f"Bad message {message!r}")
                return

            future: asyncio.Future = self.futures.pop(message.correlation_id, None)
            if future:
                try:
                    future.set_result(ujson.loads(message.body))
                except Exception as fe:
                    logger.exception(f"Failed to process response as json: {fe}")
                    future.set_result({})
            else:
                logger.error(
                    f"Failed to handle response: unknown correlation_id\nmessage: {message}\nbody:{message.body}\nfutures:{self.futures}")
        except Exception as e:
            logger.exception(
                f"Failed to handle response: {e}\nmessage: {message}\nbody:{message.body}\nfutures:{self.futures}")

    async def SendRPCDictMessage(self, queue: str, body: dict) -> dict:
        #logger.debug(f"Sending RPC msg: {body}")
        return await self.SendRPCMessage(queue=queue, body=ujson.dumps(body).encode())

    async def ReplyMessage(self, response: bytes, message: aio_pika.abc.AbstractIncomingMessage):
        try:
            connection = await self.GetConnection()
            async with connection.channel(on_return_raises=True) as chan:
                exchange = chan.default_exchange
                newMessage = aio_pika.Message(
                    body=response,
                    content_type="application/json",
                    correlation_id=message.correlation_id)

                await exchange.publish(
                    newMessage,
                    routing_key=message.reply_to,
                    mandatory=False
                )

        except Exception as e:
            logger.exception(f"[-] failed to send reply message: {e}")
            pass

    async def ReceiveFromMythicDirectExchange(self, queue: str, routing_key: str,
                                              handler: Coroutine[any, any, None]):
        while True:
            try:
                connection = await self.GetConnection()
                async with connection.channel() as chan:
                    exchange = await chan.declare_exchange(
                        name="mythic_exchange",
                        type="direct",
                        durable=True,
                        auto_delete=True,
                        internal=False,
                    )
                    q = await chan.declare_queue(
                        name=queue,
                        durable=False,
                        auto_delete=True,
                        exclusive=False,
                    )
                    await q.bind(
                        exchange=exchange,
                        routing_key=routing_key,
                    )
                    await q.consume(
                        callback=partial(directExchangeCallback, trueFunction=handler)
                    )
                    logger.info(f"[*] started listening for messages on {queue}")
                    try:
                        await asyncio.Future()
                        logger.error(f"asyncio.Future() finished in ReceiveFromMythicDirectExchange for {queue}")
                    except Exception as directException:
                        logger.exception(f"[-] exception trying to listen for direct messages on {queue}\n{directException}")
            except aiormq.exceptions.ChannelLockedResource:
                logger.error(f"[-] Another instance of this service, {queue.split('_')[0]}, is running, failed to start, trying again...")
                await asyncio.sleep(failedConnectRetryDelay)
            except Exception as e:
                logger.exception(f"[-] stopped listening for messages on {queue}, {e}")
                await asyncio.sleep(failedConnectRetryDelay)

    async def ReceiveFromRPCQueue(self, queue: str, routing_key: str, handler: Coroutine[any, any, None]):
        while True:
            try:
                connection = await self.GetConnection()
                async with connection.channel() as chan:
                    exchange = await chan.declare_exchange(
                        name="mythic_exchange",
                        type="direct",
                        durable=True,
                        auto_delete=True,
                        internal=False,

                    )
                    q = await chan.declare_queue(
                        name=queue,
                        durable=False,
                        auto_delete=True,
                        exclusive=True,
                    )
                    await q.bind(
                        exchange=exchange,
                        routing_key=routing_key,
                    )
                    await q.consume(
                        callback=partial(rpcExchangeCallback, trueFunction=handler)
                    )
                    logger.info(f"[*] started listening for messages on {queue}")
                    try:
                        await asyncio.Future()
                    finally:
                        logger.error(f"asyncio.Future() finished in ReceiveFromRPCQueue for queue {queue}")
            except aiormq.exceptions.ChannelLockedResource:
                logger.error(f"[-] Another instance of this service, {queue.split('_')[0]}, is running, failed to start, trying again...")
                await asyncio.sleep(failedConnectRetryDelay)
            except Exception as e:
                logger.exception(f"[-] stopped listening for messages on {queue}, {e}")
                await asyncio.sleep(failedConnectRetryDelay)

    async def ReceiveFromMythicDirectTopicExchange(self, queue: str, routing_key: str,
                                                   handler: Coroutine[any, any, None]):
        while True:
            try:
                connection = await self.GetConnection()
                async with connection.channel() as chan:
                    exchange = await chan.declare_exchange(
                        name="mythic_topic_exchange",
                        type="topic",
                        durable=True,
                        auto_delete=True,
                        internal=False,
                    )
                    q = await chan.declare_queue(
                        name="",
                        durable=False,
                        auto_delete=True,
                        exclusive=False,
                    )
                    await q.bind(
                        exchange=exchange,
                        routing_key=queue,
                    )
                    await q.consume(
                        callback=partial(directExchangeCallback, trueFunction=handler)
                    )
                    logger.info(f"[*] started listening for messages on {queue}")
                    try:
                        await asyncio.Future()
                        logger.error(f"asyncio.Future() finished in ReceiveFromMythicDirectTopicExchange for {queue}")
                    except Exception as directException:
                        logger.exception(f"[-] exception trying to listen for direct messages on {queue}\n{directException}")
            except Exception as e:
                logger.exception(f"[-] stopped listening for messages on {queue}, {e}")
                await asyncio.sleep(failedConnectRetryDelay)



import asyncio
from abc import abstractmethod
from .grpc.translationContainerGRPC_pb2_grpc import TranslationContainerStub
from .grpc import translationContainerGRPC_pb2 as grpcFuncs
import grpc.aio
from mythic_container.logging import logger
from .config import settings
import json
from typing import List


class TrGenerateEncryptionKeysMessage:
    """
    Message from gRPC asking translation container to generate encryption/decryption keys

    :param TranslationContainerName
    :param C2Name
    :param CryptoParamValue - what the user selected when building an agent
    :param CryptoParamName - the name of the parameter that needs a crypto key generated

    """
    def __init__(self,
                 TranslationContainerName: str,
                 C2Name: str,
                 CryptoParamValue: str,
                 CryptoParamName: str):
        self.TranslationContainerName = TranslationContainerName
        self.C2Name = C2Name
        self.CryptoParamValue = CryptoParamValue
        self.CryptoParamName = CryptoParamName

    def to_json(self):
        return {
            "translation_container_name": self.TranslationContainerName,
            "c2_profile_name": self.C2Name,
            "value": self.CryptoParamValue,
            "name": self.CryptoParamName
        }


class TrGenerateEncryptionKeysMessageResponse:
    """
    Response to a request to generate encryption keys

    :param Success indicate if this was successful or not
    :param Error if this wasn't successful, what's the error message
    :param EncryptionKey the bytes of the new encryption key to use
    :param DecryptionKey the bytes of the new decryption key to use

    """
    def __init__(self,
                 Success: bool = False,
                 Error: str = "",
                 EncryptionKey: bytes = None,
                 DecryptionKey: bytes = None):
        self.Success = Success
        self.Error = Error
        self.EncryptionKey = EncryptionKey
        self.DecryptionKey = DecryptionKey

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "enc_key": self.EncryptionKey,
            "dec_key": self.DecryptionKey
        }


class CryptoKeys:
    """
    Crypto Key information available for a translation container to decrypt/encrypt a message if necessary

    :param EncKey the bytes of the encryption key or None
    :param DecKey the bytes of the decryption key or None
    :param Value the value selected by the user for the type of encryption
    """
    def __init__(self,
                 EncKey: bytes = None,
                 DecKey: bytes = None,
                 Value: str = ""):
        self.EncKey = EncKey
        self.DecKey = DecKey
        self.Value = Value
    def to_json(self):
        return {
            "enc_key": self.EncKey,
            "dec_key": self.DecKey,
            "value": self.Value
        }


class TrCustomMessageToMythicC2FormatMessage:
    """
    Request to translation a custom C2 formatted message into a Mythic dictionary message

    :param TranslationContainerName
    :param C2Name
    :param Message the raw message to be translated WITHOUT the UUID in front
    :param UUID the extracted UUID from the front of the message
    :param MythicEncrypts indicates if Mythic will do the encryption/decryption or not
    :param CryptoKeys the encryption/decryption keys Mythic is tracking in case the translation container needs them
    """
    def __init__(self,
                 TranslationContainerName: str,
                 C2Name: str,
                 Message: bytes,
                 UUID: str,
                 MythicEncrypts: bool,
                 CryptoKeys: List[CryptoKeys] = []):
        self.TranslationContainerName = TranslationContainerName
        self.C2Name = C2Name
        self.Message = Message
        self.UUID = UUID
        self.MythicEncrypts = MythicEncrypts
        self.CryptoKeys = CryptoKeys

    def to_json(self):
        return {
            "translation_container_name": self.TranslationContainerName,
            "c2_profile_name": self.C2Name,
            "message": self.Message,
            "uuid": self.UUID,
            "mythic_encrypts": self.MythicEncrypts,
            "crypto_keys": [x.to_json() for x in self.CryptoKeys]
        }


class TrCustomMessageToMythicC2FormatMessageResponse:
    """
    Response to a request to translate a custom C2 formatted message into Mythic dictionary format

    :param Success indicating if this translation was successful or not
    :param Error if success is false, what is the error string to report back to the operator
    :param Message the Mythic dictionary message
    """
    def __init__(self,
                 Success: bool = False,
                 Error: str = "",
                 Message: dict = {}):
        self.Success = Success
        self.Error = Error
        self.Message = Message

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message
        }


class TrMythicC2ToCustomMessageFormatMessage:
    """
    Request to translate a Mythic message into a custom C2 format

    :param TranslationContainerName
    :param C2Name
    :param Message the Mythic dictionary message
    :param UUID the UUID associated with this message
    :param MythicEncrypts does Mythic handle encryption or not, if not, then this function needs to encrypt
    :param CryptoKeys the encryption/decryption keys that Mythic is tracking for this UUID
    """
    def __init__(self,
                 TranslationContainerName: str,
                 C2Name: str,
                 Message: dict,
                 UUID: str,
                 MythicEncrypts: bool,
                 CryptoKeys: List[CryptoKeys]):
        self.TranslationContainerName = TranslationContainerName
        self.C2Name = C2Name
        self.Message = Message
        self.UUID = UUID
        self.MythicEncrypts = MythicEncrypts
        self.CryptoKeys = CryptoKeys

    def to_json(self):
        return {
            "translation_container_name": self.TranslationContainerName,
            "c2_profile_name": self.C2Name,
            "message": self.Message,
            "uuid": self.UUID,
            "mythic_encrypts": self.MythicEncrypts,
            "crypto_keys": [x.to_json() for x in self.CryptoKeys]
        }


class TrMythicC2ToCustomMessageFormatMessageResponse:
    """
    Response to a request for converting a Mythic dictionary message to a custom C2 format

    :param Success was this translation successful or not
    :param Error if this was not successful, what was the error
    :param Message The final message that should be returned back to the agent. No other processing happens to this message after this function, so add the UUID as needed and base64 encode it.
    """
    def __init__(self,
                 Success: bool = False,
                 Error: str = "",
                 Message: bytes = b''):
        self.Success = Success
        self.Error = Error
        self.Message = Message

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message
        }


class TranslationContainer:
    """The base definition for a Translation Container Service.

    To have this hooked up to a Payload Type, you need to specify this service's name as the translation_container attribute in your Payload Type.

    This uses gRPC to connect to port 17444 on the Mythic Server.

    Attributes:
        name (str): The name of the translation container
        description (str): Description of the translation container to appear in Mythic's UI
        author (str): The author of the container

    Functions:
        to_json:
            return dictionary form of class
        generate_keys:
            A function for generating encryption/decryption keys based on provided C2 info
        translate_to_c2_format:
            A function for translating from Mythic's JSON format to a custom C2 format
        translate_from_c2_format:
            A function for translating from a custom C2 format to Mythic's JSON format

    """
    async def generate_keys(self, inputMsg: TrGenerateEncryptionKeysMessage) -> TrGenerateEncryptionKeysMessageResponse:
        response = TrGenerateEncryptionKeysMessageResponse(Success=False)
        response.Error = f"Not Implemented:\n{inputMsg}"

        return response

    async def translate_to_c2_format(self,
                                     inputMsg: TrMythicC2ToCustomMessageFormatMessage) -> TrMythicC2ToCustomMessageFormatMessageResponse:
        response = TrMythicC2ToCustomMessageFormatMessageResponse(Success=False)
        response.Error = f"Not Implemented:\n{inputMsg}"
        return response

    async def translate_from_c2_format(self,
                                       inputMsg: TrCustomMessageToMythicC2FormatMessage) -> TrCustomMessageToMythicC2FormatMessageResponse:
        response = TrCustomMessageToMythicC2FormatMessageResponse(Success=False)
        response.Error = f"Not Implemented:\n{inputMsg}"
        return response

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def description(self):
        pass

    @property
    @abstractmethod
    def author(self):
        pass

    def to_json(self):
        return {
            "name": self.name,
            "description": self.description,
            "author": self.author,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


translationServices: dict[str, TranslationContainer] = {}


async def handleTranslationServices(tr_name: str):
    while True:
        try:
            logger.info(f"Attempting connection to gRPC for {tr_name}...")
            channel = grpc.aio.insecure_channel(
                f'{settings.get("mythic_server_host", "127.0.0.1")}:{settings.get("mythic_server_grpc_port", 17444)}')
            await channel.channel_ready()
            client = TranslationContainerStub(channel=channel)
            genKeys = handleGenerateKeys(tr_name, client)
            customToMythic = handleCustomToMythic(tr_name, client)
            mythicToCustom = handleMythicToCustom(tr_name, client)
            logger.info(f"[+] Successfully connected to gRPC for {tr_name}")
            await asyncio.gather(genKeys, customToMythic, mythicToCustom)
        except Exception as e:
            logger.exception(f"Translation gRPC services closed for {tr_name}: {e}")


async def handleGenerateKeys(tr_name: str, client):
    try:
        while True:
            stream = client.GenerateEncryptionKeys()
            await stream.write(grpcFuncs.TrGenerateEncryptionKeysMessageResponse(
                Success=True,
                TranslationContainerName=tr_name
            ))
            logger.info(f"Connected to gRPC for generating encryption keys for {tr_name}")
            async for request in stream:
                try:
                    result = await translationServices[tr_name].generate_keys(TrGenerateEncryptionKeysMessage(
                        TranslationContainerName=request.TranslationContainerName,
                        C2Name=request.C2Name,
                        CryptoParamName=request.CryptoParamName,
                        CryptoParamValue=request.CryptoParamValue
                    ))
                    await stream.write(grpcFuncs.TrGenerateEncryptionKeysMessageResponse(
                        Success=result.Success,
                        TranslationContainerName=tr_name,
                        Error=result.Error,
                        EncryptionKey=result.EncryptionKey,
                        DecryptionKey=result.DecryptionKey
                    ))
                except Exception as d:
                    logger.exception(f"Failed to process handleGenerateKeys message:\n{d}")
                    await stream.write(grpcFuncs.TrGenerateEncryptionKeysMessageResponse(
                        Success=False,
                        TranslationContainerName=tr_name,
                        Error=f"Failed to process handleGenerateKeys message:\n{d}"
                    ))
            logger.error(f"disconnected from gRPC for generating encryption keys for {tr_name}")
    except Exception as e:
        logger.exception(f"[-] exception in handleGenerateKeys for {tr_name}")


async def handleCustomToMythic(tr_name: str, client):
    try:
        while True:
            stream = client.TranslateFromCustomToMythicFormat()
            await stream.write(grpcFuncs.TrCustomMessageToMythicC2FormatMessageResponse(
                Success=True,
                TranslationContainerName=tr_name
            ))
            logger.info(f"Connected to gRPC for handling CustomC2 to MythicC2 Translations for {tr_name}")
            async for request in stream:
                try:
                    grpcCryptoKeys = request.CryptoKeys
                    localCryptoKeys = []
                    for keys in grpcCryptoKeys:
                        localCryptoKeys.append(CryptoKeys(
                            Value=keys.Value,
                            DecKey=keys.DecKey,
                            EncKey=keys.EncKey
                        ))
                    inputMsg = TrCustomMessageToMythicC2FormatMessage(
                        TranslationContainerName=request.TranslationContainerName,
                        C2Name=request.C2Name,
                        Message=request.Message,
                        UUID=request.UUID,
                        MythicEncrypts=request.MythicEncrypts,
                        CryptoKeys=localCryptoKeys
                    )
                    result = await translationServices[tr_name].translate_from_c2_format(inputMsg)
                    response = grpcFuncs.TrCustomMessageToMythicC2FormatMessageResponse(
                        Success=result.Success,
                        Error=result.Error,
                        TranslationContainerName=tr_name,
                        Message=json.dumps(result.Message).encode()
                    )
                    await stream.write(response)
                except Exception as d:
                    logger.exception(f"Failed to process handleCustomToMythic message:\n{d}")
                    await stream.write(grpcFuncs.TrCustomMessageToMythicC2FormatMessageResponse(
                        Success=False,
                        TranslationContainerName=tr_name,
                        Error=f"Failed to process handleCustomToMythic message:\n{d}"
                    ))
            logger.error(f"disconnected from gRPC for doing custom->mythic c2 for {tr_name}")
    except Exception as e:
        logger.exception(f"[-] exception in handleCustomToMythic for {tr_name}")


async def handleMythicToCustom(tr_name: str, client):
    try:
        while True:
            stream = client.TranslateFromMythicToCustomFormat()
            await stream.write(grpcFuncs.TrMythicC2ToCustomMessageFormatMessageResponse(
                Success=True,
                TranslationContainerName=tr_name
            ))
            logger.info(f"Connected to gRPC for handling MythicC2 to CustomC2 Translations for {tr_name}")
            async for request in stream:
                try:
                    grpcCryptoKeys = request.CryptoKeys
                    localCryptoKeys = []
                    for keys in grpcCryptoKeys:
                        localCryptoKeys.append(CryptoKeys(
                            Value=keys.Value,
                            DecKey=keys.DecKey,
                            EncKey=keys.EncKey
                        ))
                    inputMsg = TrMythicC2ToCustomMessageFormatMessage(
                        TranslationContainerName=request.TranslationContainerName,
                        C2Name=request.C2Name,
                        UUID=request.UUID,
                        MythicEncrypts=request.MythicEncrypts,
                        CryptoKeys=localCryptoKeys,
                        Message=json.loads(request.Message)
                    )
                    result = await translationServices[tr_name].translate_to_c2_format(inputMsg)
                    response = grpcFuncs.TrMythicC2ToCustomMessageFormatMessageResponse(
                        Success=result.Success,
                        Error=result.Error,
                        TranslationContainerName=tr_name,
                        Message=result.Message
                    )
                    await stream.write(response)
                except Exception as d:
                    logger.exception(f"Failed to process handleMythicToCustom message:\n{d}")
                    await stream.write(grpcFuncs.TrMythicC2ToCustomMessageFormatMessageResponse(
                        Success=False,
                        TranslationContainerName=tr_name,
                        Error=f"Failed to process handleMythicToCustom message:\n{d}"
                    ))
            logger.error(f"disconnected from gRPC for doing mythic->custom c2 for {tr_name}")
    except Exception as e:
        logger.exception(f"[-] exception in handleMythicToCustom for {tr_name}")



import pathlib
from enum import Enum
from abc import abstractmethod
import json
from collections.abc import Awaitable, Callable
from .logging import logger
import base64
import sys
from typing import List


class C2OPSECMessage:
    """Payload's C2 Profile configuration for OPSEC checking

    Attributes:
        Name (str): Name of the C2 Profile
        Parameters (dict): Dictionary of C2 Parameter name:value

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 parameters: dict,
                 **kwargs):
        self.Name = c2_profile_name
        self.Parameters = parameters
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self) -> dict:
        return {
            "c2_profile_name": self.Name,
            "parameters": self.Parameters
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2OPSECMessageResponse:
    """Status of running C2 Profile's opsec function

    Attributes:
        Success (bool): Did the OPSEC check succeed or fail
        Error (str): Error message if the OPSEC check failed
        Message (str): Informative message about the OPSEC check if there was no error

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2ConfigCheckMessage:
    """Payload's C2 Profile configuration for Configuration checking

    Attributes:
        Name (str): Name of the C2 Profile
        Parameters (dict): Dictionary of C2 Parameter name:value

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 parameters: dict,
                 **kwargs):
        self.Name = c2_profile_name
        self.Parameters = parameters
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "parameters": self.Parameters
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2ConfigCheckMessageResponse:
    """Status of running C2 Profile's config_check function

    Attributes:
        Success (bool): Did the config check succeed or fail
        Error (str): Error message if the config check failed
        Message (str): Informative message about the config check if there was no error

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetIOCMessage:
    """Payload's C2 Profile configuration for getting IOCs

    Attributes:
        Name (str): Name of the C2 Profile
        Parameters (dict): Dictionary of C2 Parameter name:value

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 parameters: dict,
                 **kwargs):
        self.Name = c2_profile_name
        self.Parameters = parameters
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "parameters": self.Parameters
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetIOCMessageResponseIOC:
    """An IOC to report back as part of a C2 Profile's get_ioc function

    Attributes:
        Type (str): The type of IOC (ex: Hash, URL)
        IOC (str): The actual IOC value
    """
    def __init__(self,
                 Type: str,
                 IOC: str):
        self.Type = Type
        self.IOC = IOC

    def to_json(self):
        return {
            "type": self.Type,
            "ioc": self.IOC
        }


class C2GetIOCMessageResponse:
    """Status of running C2 Profile's get_ioc function

    Attributes:
        Success (bool): Did the get ioc request succeed or fail
        Error (str): Error message if the get ioc request failed
        IOCs (List[C2GetIOCMessageResponseIOC]): List of the IOC values

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 IOCs: List[C2GetIOCMessageResponseIOC] = [],
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.IOCs = IOCs
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "iocs": self.IOCs
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2SampleMessageMessage:
    """Sample C2 message based on the Payload's configuration

    Attributes:
        Name (str): Name of the C2 Profile
        Parameters (dict): Dictionary of C2 Parameter name:value

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 parameters: dict,
                 **kwargs):
        self.Name = c2_profile_name
        self.Parameters = parameters
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "parameters": self.Parameters
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2SampleMessageMessageResponse:
    """Sample C2 Message based on the payload's configuration

    Attributes:
        Success (bool): Did the sample message generation succeed or fail
        Error (str): Error message if the sample message generation failed
        Message (str): Sample message to represent that C2 traffic would look like for this payload

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetDebugOutputMessage:
    """Try to get debug output from subprocess running internal server code

    Attributes:
        Name (str): Name of the C2 Profile

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 **kwargs):
        self.Name = c2_profile_name
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetDebugOutputMessageResponse:
    """Debug output from subprocesses running internal server

    Attributes:
        Success (bool): Were we able to get information from subprocess
        Error (str): Error message if the subprocess exited
        Message (str): Whatever output was generated from the subprocess
        InternalServerRunning (bool): Is the subprocess still running or not

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 InternalServerRunning: bool = False,
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        self.InternalServerRunning = InternalServerRunning
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message,
            "server_running": self.InternalServerRunning
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetFileMessage:
    """Fetch file from C2 Profile's server folder directory

    Attributes:
        Name (str): Name of the C2 Profile
        Filename (str): Name of the file to read

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 filename: str,
                 **kwargs):
        self.Name = c2_profile_name
        self.Filename = filename
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "filename": self.Filename
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetFileMessageResponse:
    """Contents or error from fetching a file

    Attributes:
        Success (bool): Did we successfully read the file or not
        Error (str): Error message if we failed to read the file
        Message (str): Contents of the file

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: bytes = b"",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": base64.b64encode(self.Message).decode(),
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetRedirectorRulesMessage:
    """Payload's C2 Profile configuration for generating Apache ModRewrite rules

    Attributes:
        Name (str): Name of the C2 Profile
        Parameters (dict): Dictionary of C2 Parameter name:value

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 parameters: dict,
                 **kwargs):
        self.Name = c2_profile_name
        self.Parameters = parameters
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "parameters": self.Parameters
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2GetRedirectorRulesMessageResponse:
    """Status of running C2 Profile's get redirector rules function

    Attributes:
        Success (bool): Did the redirector check succeed or fail
        Error (str): Error message if the redirector check failed
        Message (str): Apache ModRewrite rules

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2ListFileMessage:
    """List the contents of a C2 Profile's server folder directory

    Attributes:
        Name (str): Name of the C2 Profile

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 **kwargs):
        self.Name = c2_profile_name
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2ListFileMessageResponse:
    """List of filenames in the C2 Profile's server folder directory

    Attributes:
        Success (bool): Did the OPSEC check succeed or fail
        Error (str): Error message if the OPSEC check failed
        Files (list[str]): List of filenames

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Files: list[str] = [],
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Files = Files
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "files": self.Files,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2RemoveFileMessage:
    """Request to remove a file from a C2 Profile's server folder path

    Attributes:
        Name (str): Name of the C2 Profile
        Filename (str): Name of the file to remove

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 filename: str,
                 **kwargs):
        self.Name = c2_profile_name
        self.Filename = filename
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "filename": self.Filename
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2RemoveFileMessageResponse:
    """Status of removing the file

    Attributes:
        Success (bool): Did the file removal succeed or fail
        Error (str): Error message if the file removal failed

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2StartServerMessage:
    """Request to start a C2 Profile's internal server as a subprocess

    Attributes:
        Name (str): Name of the C2 Profile

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 **kwargs):
        self.Name = c2_profile_name
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2StartServerMessageResponse:
    """Status of running C2 Profile's internal server as a subprocess

    Attributes:
        Success (bool): Did the start subprocess succeed or fail
        Error (str): Error message if the subprocess failed to run for more than 3 seconds
        Message (str): Output from the first 3 seconds of the subprocess running
        InternalServerRunning (bool): Is the internal server subprocess still running or not

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 InternalServerRunning: bool = False,
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        self.InternalServerRunning = InternalServerRunning
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message,
            "server_running": self.InternalServerRunning
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2StopServerMessage:
    """Request to stop the C2 Profile's internal server subprocess

    Attributes:
        Name (str): Name of the C2 Profile

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 **kwargs):
        self.Name = c2_profile_name
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2StopServerMessageResponse:
    """Status of stopping the C2 Server's subprocess

    Attributes:
        Success (bool): Did the subprocesses successfully stop or not
        Error (str): Error message if the subprocess failed to stop
        Message (str): Any final output the subprocess was reporting
        InternalServerRunning (bool): Is the internal subprocess still running or not

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: str = "",
                 InternalServerRunning: bool = False,
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        self.InternalServerRunning = InternalServerRunning
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message,
            "server_running": self.InternalServerRunning
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2WriteFileMessage:
    """Request to write a file to the C2 Profile's server folder directory

    Attributes:
        Name (str): Name of the C2 Profile
        Filename (str): Name of the file to create/overwrite
        Contents (bytes): Contents of the file to write

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 c2_profile_name: str,
                 filename: str,
                 contents: str,
                 **kwargs):
        self.Name = c2_profile_name
        self.Filename = filename
        self.Contents = base64.b64decode(contents)
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "c2_profile_name": self.Name,
            "filename": self.Filename,
            "contents": base64.b64encode(self.Contents)
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2WriteFileMessageResponse:
    """Status of writing the file to disk

    Attributes:
        Success (bool): Did the file get written or not
        Error (str): Error message if the file failed to write to disk
        Message (str): Informative message about the file getting written

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 Success: bool,
                 Error: str = "",
                 Message: bytes = b"",
                 **kwargs):
        self.Success = Success
        self.Error = Error
        self.Message = Message
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "message": self.Message,
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2OtherServiceRPCMessage:
    """Request to call an RPC function of another C2 Profile or Payload Type

    Attributes:
        ServiceName (str): Name of the C2 Profile or Payload Type
        ServiceRPCFunction (str): Name of the function to call
        ServiceRPCFunctionArguments (dict): Arguments to that function

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 ServiceName: str = None,
                 service_name: str = None,
                 ServiceRPCFunction: str = None,
                 service_function: str = None,
                 ServiceRPCFunctionArguments: dict = None,
                 service_arguments: dict = None,
                 **kwargs):
        self.ServiceName = ServiceName
        if self.ServiceName is None:
            self.ServiceName = service_name
        self.ServiceRPCFunction = ServiceRPCFunction
        if self.ServiceRPCFunction is None:
            self.ServiceRPCFunction = service_function
        self.ServiceRPCFunctionArguments = ServiceRPCFunctionArguments
        if self.ServiceRPCFunctionArguments is None:
            self.ServiceRPCFunctionArguments = service_arguments
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "service_name": self.ServiceName,
            "service_function": self.ServiceRPCFunction,
            "service_arguments": self.ServiceRPCFunctionArguments
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2OtherServiceRPCMessageResponse:
    """Result of running an RPC call from another service

    Attributes:
        Success (bool): Did the RPC succeed or fail
        Error (str): Error message if the RPC check failed
        Result (dict): Result from the RPC

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(self,
                 success: bool = None,
                 error: str = None,
                 result: dict = None,
                 Success: bool = None,
                 Error: str = None,
                 Result: dict = None,
                 **kwargs):
        self.Success = Success
        if self.Success is None:
            self.Success = success
        self.Error = Error
        if self.Error is None:
            self.Error = error
        self.Result = Result
        if self.Result is None:
            self.Result = result
        for k, v in kwargs.items():
            logger.error(f"unknown kwarg {k} {v}")

    def to_json(self):
        return {
            "success": self.Success,
            "error": self.Error,
            "result": self.Result
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class ParameterType(Enum):
    """Types of parameters available for C2 Profiles when building payloads

     Attributes:
         String:
             A string value
         ChooseOne:
             A list of choices for the user to select exactly one
         Array:
             The user can supply multiple values in an Array format
         Date:
             The user can select a Date in YYYY-MM-DD format
         Dictionary:
             The user can supply a dictionary of values
         Boolean:
             The user can toggle a switch for True/False
     """
    String = "String"
    ChooseOne = "ChooseOne"
    Array = "Array"
    Date = "Date"
    Dictionary = "Dictionary"
    Boolean = "Boolean"


class DictionaryChoice:
    """A single dictionary choice/option for the UI when select C2 Profile Parameters

    Attributes:
        name (str): Name of the choice
        default_show (bool): Should this be pre-selected with a default value
        default_value (str): Default value to fill in

    """

    def __init__(self,
                 name: str,
                 default_value: str = "",
                 default_show: bool = True):
        self.name = name
        self.default_show = default_show
        self.default_value = default_value

    def to_json(self):
        return {
            "name": self.name,
            "default_value": self.default_value,
            "default_show": self.default_show
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2ProfileParameter:
    """C2Profile Parameter Definition for use when generating payloads

    Attributes:
        name (str): Name of the parameter for scripting and for when building payloads
        description (str): Informative description displayed when building a payload
        default_value (any): Default value to pre-populate
        randomize (bool): Should this value be randomized (requires format_string)
        format_string (str): A regex used for randomizing values if randomize is true
        parameter_type (ParameterType): The type of parameter this is
        required (bool): Is this parameter required to have a non-empty value or not
        verifier_regex (str): Regex used to verify that the user typed something appropriate
        choices (list[str]): Choices for ChooseOne parameter type
        dictionary_choices (list[DictionaryChoice]): Configuration options for the Dictionary parameter type
        crypto_type (bool): Indicate if this value should be used to generate a crypto key or not

    Functions:
        to_json(self): return dictionary form of class
    """

    def __init__(
            self,
            name: str,
            description: str,
            default_value: any = None,
            randomize: bool = False,
            format_string: str = "",
            parameter_type: ParameterType = ParameterType.String,
            required: bool = True,
            verifier_regex: str = "",
            choices: list[str] = None,
            dictionary_choices: list[DictionaryChoice] = None,
            crypto_type: bool = False,
    ):
        self.name = name
        self.description = description
        self.randomize = randomize
        self.format_string = format_string
        self.parameter_type = parameter_type
        self.required = required
        self.verifier_regex = verifier_regex
        self.choices = choices
        self.default_value = default_value
        self.crypto_type = crypto_type
        self.dictionary_choices = dictionary_choices

    def to_json(self):
        return {
            "name": self.name,
            "description": self.description,
            "default_value": self.default_value,
            "randomize": self.randomize,
            "format_string": self.format_string,
            "required": self.required,
            "parameter_type": self.parameter_type.value,
            "verifier_regex": self.verifier_regex,
            "crypto_type": self.crypto_type,
            "choices": self.choices,
            "dictionary_choices": [x.to_json() for x in
                                   self.dictionary_choices] if self.dictionary_choices is not None else None
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


class C2Profile:
    """C2 Profile to import and sync with Mythic

    Attributes:
        name (str): Name of the profile
        description (str): Description of the profile
        author (str): Author of the profile
        is_p2p (bool): Is the profile defining P2P communications or egress communications
        parameters (list[C2ProfileParameter]): Definition of the parameters for the profile used when building a payload
        server_folder_path (pathlib.Path): Path to the folder that contains `config.json`
        server_binary_path (pathlib.Path): Path to the binary that gets executed when "Start" is clicked in the UI
        custom_rpc_functions: Dictionary of RPC name to awaitable RPC function that other services can call

    Functions:
        opsec

        config_check

        redirect_rules

        get_ioc

        sample_message

    """

    async def opsec(self, inputMsg: C2OPSECMessage) -> C2OPSECMessageResponse:
        """Check payload's C2 configuration for OPSEC issues

        :param inputMsg: Payload's C2 Profile configuration
        :return: C2OPSECMessageResponse detailing the results of the OPSEC check
        """
        response = C2OPSECMessageResponse(Success=True)
        response.Message = "Not Implemented, passing by default"
        response.Message += f"\nInput: {json.dumps(inputMsg.to_json(), indent=4)}"
        return response

    async def config_check(self, inputMsg: C2ConfigCheckMessage) -> C2ConfigCheckMessageResponse:
        """Check a payload's C2 configuration to see if it matches the local configuration

        :param inputMsg: Payload's C2 Profile configuration
        :return: C2ConfigCheckMessageResponse detailing the results of the configuration check
        """
        response = C2ConfigCheckMessageResponse(Success=True)
        response.Message = "Not Implemented"
        response.Message += f"\nInput: {json.dumps(inputMsg.to_json(), indent=4)}"
        return response

    async def redirect_rules(self, inputMsg: C2GetRedirectorRulesMessage) -> C2GetRedirectorRulesMessageResponse:
        """Generate Apache ModRewrite rules given the Payload's C2 configuration

        :param inputMsg: Payload's C2 Profile configuration
        :return: C2GetRedirectorRulesMessageResponse detailing some Apache ModRewrite rules for the payload
        """
        response = C2GetRedirectorRulesMessageResponse(Success=True)
        response.Message = "Not Implemented"
        response.Message += f"\nInput: {json.dumps(inputMsg.to_json(), indent=4)}"
        return response

    async def get_ioc(self, inputMsg: C2GetIOCMessage) -> C2GetIOCMessageResponse:
        """Generate IOCs for the network traffic associated with the specified c2 configuration

        :param inputMsg: Payload's C2 Profile configuration
        :return: C2GetIOCMessageResponse detailing some IOCs
        """
        response = C2GetIOCMessageResponse(Success=True)
        response.IOCs = []
        return response

    async def sample_message(self, inputMsg: C2SampleMessageMessage) -> C2SampleMessageMessageResponse:
        """Generate a sample message for this c2 profile based on the configuration specified

        :param inputMsg: Payload's C2 Profile configuration
        :return: C2SampleMessageMessageResponse detailing a sample message
        """
        response = C2SampleMessageMessageResponse(Success=True)
        response.Message = "Not Implemented"
        response.Message += f"\nInput: {json.dumps(inputMsg.to_json(), indent=4)}"
        return response

    custom_rpc_functions: dict[
        str, Callable[[C2OtherServiceRPCMessage], Awaitable[C2OtherServiceRPCMessageResponse]]] = {}
    server_folder_path: pathlib.Path
    server_binary_path: pathlib.Path

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def description(self):
        pass

    @property
    @abstractmethod
    def author(self):
        pass

    @property
    @abstractmethod
    def is_p2p(self):
        pass

    @property
    @abstractmethod
    def is_server_routed(self):
        pass

    @property
    @abstractmethod
    def parameters(self):
        pass

    def to_json(self):
        if self.server_binary_path is None:
            logger.exception(
                "Must supply server_binary_path as pathlib.Path pointing to the binary to run when the server starts")
            sys.exit(1)
        elif not isinstance(self.server_binary_path, pathlib.Path):
            logger.exception("Wrong type for server_binary_path - should be pathlib.Path")
            sys.exit(1)
        elif self.server_folder_path is None:
            logger.exception(
                "Must supply server_folder_path as pathlib.Path pointing to folder where your c2 files exist")
            sys.exit(1)
        elif not isinstance(self.server_folder_path, pathlib.Path):
            logger.exception("Wrong type for server_folder_path - should be pathlib.Path")
            sys.exit(1)
        return {
            "c2_profile": {
                "name": self.name,
                "description": self.description,
                "author": self.author,
                "is_p2p": self.is_p2p,
                "is_server_routed": self.is_server_routed,
            },
            "parameters": [x.to_json() for x in self.parameters]
        }

    def __str__(self):
        return json.dumps(self.to_json(), sort_keys=True, indent=2)


c2Profiles: dict[str, C2Profile] = {}
runningServers: dict[str, dict] = {}
